use ExtUtils::MakeMaker;

my ($define, $ccfl);

# special flag for AIX
($^O =~ m/AIX/i) ? ($ccfl = '-qcpluscmt') : ($ccfl = '');

# path libsmbclient.h
my $include = try_to_find("libsmbclient.h");
$include = prompt("Where can I find libsmbclient.h ?", $include);
warn_user("$include/libsmbclient.h") if (!-r "$include/libsmbclient.h");

# path libsmbclient.so
my $lib = try_to_find("libsmbclient.so");
$lib = prompt("Where can I find libsmbclient.so ?",$lib);
warn_user("libsmbclient.so") if (!-r "$lib/libsmbclient.so");

# tests demande ?
my $ans = 
  prompt("Do you want make tests ?(you will be prompted for server / \n".
	   "worgroup / share / user / password to make read write access.",
	   "no");
if ($ans =~ /^y(es)?$/i) 
  {
    my $server = prompt("Server ?","localhost");
    my $share = prompt("Share ?","homes");
    my $workgroup = prompt("Workgroup/Domain ?","");
    my $user = prompt("Smb user ?",$ENV{'USER'});
    my $pass = prompt("Smb password ?");
    open(FILE,">.c") || warn "Can't create .c for test: $!\n";
    print FILE $server,"\t",$share,"\t",$workgroup,"\t",$user,"\t",$pass,"\n";
    close FILE;
  }

# Trace for debug
$ans = prompt("Compile module with trace to STDERR ?", "no");
if ($ans =~ /^y(es)?$/i) { $define = "-DVERBOSE"; }

# Create config.h for alloca via configure
my $prefix = $include;
$prefix=$1 = $1 if $include=~m!(.*)include$!;
system("./configure --with-samba-prefix=$prefix") == 0
  or die "Can't run configure:$!\n";

# Create makefile
WriteMakefile
  (
   'NAME'         => 'Filesys::SmbClient',
   'VERSION_FROM' => 'lib/Filesys/SmbClient.pm',
   'INC'          => "-I$include",
   'DEFINE'       => $define,
   'CCFLAGS'      => "-Wall -Wunused -Wsign-compare -Wuninitialized",
   'LIBS'         => ["-L$lib -lsmbclient"],
   'OBJECT'       => 'libauthSamba.o SmbClient.o',
   'PREREQ_PM'    => {
		       'IO::Scalar' => 0,
		       'Test::More' => 0,
#		       perl => '5.006',
		     },
   'PMLIBDIRS'    => [ 'lib', 'lib/Filesys' ],
   'clean'        => { FILES => "*~ config.status config.log config.h config.cache .c" },
    'AUTHOR'      => 'Alain BARBET (alian@alianwebserver.com)',
    'ABSTRACT'    => 'Filesys::SmbClient - Interface for access Samba filesystem with libsmclient.so',
  );

sub warn_user
  {
    my $file = $_[0];
    warn <<EOF;
*************************************************************
** WHAT !!!!
** I Can't find $file : $!. 
*************************************************************
If you don't have this file you can download last version
of Samba on www.samba.org and do:

\$ tar zxvf samba-2.2.x.tar.gz
\$ cd samba-2.2.x/source
\$ ./configure
\$ make bin/libsmbclient.so

Then you can find libsmbclient.h in source/include directory and
libsmbclient.so in source/bin directory. You can then put them in
/usr/include and /usr/lib (or /usr/local/include and 
/usr/local/lib)

EOF

  }

sub try_to_find {
  my $name = shift;
  my @path = find_path($name);
  foreach my $d (@path) {
    if (-e "$d/$name") { return $d; }
  }
}

sub find_path {
  my $ext = ($_[0]=~m!\.h$! ? "include" : "lib");
  my @defaults = (
    qw(
      /usr/local/samba3
      /usr/local/samba
      /usr/local
      /usr
    ),
    ''  # "" will be turned into "/lib", etc.
  );
  # keep the defaults in order at the end
  my %seen = map {; ($_ => 1) } @defaults;
  my @path = (
    grep { -d $_ }           # only search dirs that actually exist
    map  { ("$_/${ext}64", "$_/$ext") }  # "/lib64", "/lib"
    (
      grep { !$seen{$_}++ }  # unique
      (
        map {
          s/\/s?bin$//;      # strip trailing /bin or /sbin
          s!//!/!g;          # remove duplicate slashes
          s!/$!!g;           # remove trailing slash
          $_;                # return
        } split(/:/, $ENV{PATH})
      ),
    ),
    @defaults
  );

  print "Searching these directories: \n", map { "  $_\n" } @path;
  return @path;
}
